@@ -83,6 +83,10 @@ var invalidHTTPClientConfigs = []struct {
 		httpClientConfigFile: "testdata/http.conf.basic-auth.too-much.bad.yaml",
 		errMsg:               "at most one of basic_auth password & password_file must be configured",
 	},
+	{
+		httpClientConfigFile: "testdata/http.conf.basic-auth.bad-username.yaml",
+		errMsg:               "at most one of basic_auth username & username_file must be configured",
+	},
 	{
 		httpClientConfigFile: "testdata/http.conf.mix-bearer-and-creds.bad.yaml",
 		errMsg:               "authorization is not compatible with bearer_token & bearer_token_file",
@@ -119,6 +123,18 @@ var invalidHTTPClientConfigs = []struct {
 		httpClientConfigFile: "testdata/http.conf.oauth2-no-token-url.bad.yaml",
 		errMsg:               "oauth2 token_url must be configured",
 	},
+	{
+		httpClientConfigFile: "testdata/http.conf.proxy-from-env.bad.yaml",
+		errMsg:               "if proxy_from_environment is configured, proxy_url must not be configured",
+	},
+	{
+		httpClientConfigFile: "testdata/http.conf.no-proxy.bad.yaml",
+		errMsg:               "if proxy_from_environment is configured, no_proxy must not be configured",
+	},
+	{
+		httpClientConfigFile: "testdata/http.conf.no-proxy-without-proxy-url.bad.yaml",
+		errMsg:               "if no_proxy is configured, proxy_url must also be configured",
+	},
 }
 
 func newTestServer(handler func(w http.ResponseWriter, r *http.Request)) (*httptest.Server, error) {
@@ -447,6 +463,50 @@ func TestNewClientFromConfig(t *testing.T) {
 	}
 }
 
+func TestProxyConfiguration(t *testing.T) {
+	testcases := map[string]struct {
+		testFn  string
+		loader  func(string) (*HTTPClientConfig, []byte, error)
+		isValid bool
+	}{
+		"good yaml": {
+			testFn:  "testdata/http.conf.proxy-headers.good.yml",
+			loader:  LoadHTTPConfigFile,
+			isValid: true,
+		},
+		"bad yaml": {
+			testFn:  "testdata/http.conf.proxy-headers.bad.yml",
+			loader:  LoadHTTPConfigFile,
+			isValid: false,
+		},
+		"good json": {
+			testFn:  "testdata/http.conf.proxy-headers.good.json",
+			loader:  loadHTTPConfigJSONFile,
+			isValid: true,
+		},
+		"bad json": {
+			testFn:  "testdata/http.conf.proxy-headers.bad.json",
+			loader:  loadHTTPConfigJSONFile,
+			isValid: false,
+		},
+	}
+
+	for name, tc := range testcases {
+		t.Run(name, func(t *testing.T) {
+			_, _, err := tc.loader(tc.testFn)
+			if tc.isValid {
+				if err != nil {
+					t.Fatalf("Error validating %s: %s", tc.testFn, err)
+				}
+			} else {
+				if err == nil {
+					t.Fatalf("Expecting error validating %s but got %s", tc.testFn, err)
+				}
+			}
+		})
+	}
+}
+
 func TestNewClientFromInvalidConfig(t *testing.T) {
 	var newClientInvalidConfig = []struct {
 		clientConfig HTTPClientConfig
@@ -718,7 +778,7 @@ func TestTLSConfigInvalidCA(t *testing.T) {
 				KeyFile:            ClientKeyNoPassPath,
 				ServerName:         "",
 				InsecureSkipVerify: false},
-			errorMessage: fmt.Sprintf("unable to read specified client cert (%s) & key (%s):", MissingCert, ClientKeyNoPassPath),
+			errorMessage: fmt.Sprintf("unable to read specified client cert (%s):", MissingCert),
 		}, {
 			configTLSConfig: TLSConfig{
 				CAFile:             "",
@@ -726,7 +786,27 @@ func TestTLSConfigInvalidCA(t *testing.T) {
 				KeyFile:            MissingKey,
 				ServerName:         "",
 				InsecureSkipVerify: false},
-			errorMessage: fmt.Sprintf("unable to read specified client cert (%s) & key (%s):", ClientCertificatePath, MissingKey),
+			errorMessage: fmt.Sprintf("unable to read specified client key (%s):", MissingKey),
+		},
+		{
+			configTLSConfig: TLSConfig{
+				CAFile:             "",
+				Cert:               readFile(t, ClientCertificatePath),
+				CertFile:           ClientCertificatePath,
+				KeyFile:            ClientKeyNoPassPath,
+				ServerName:         "",
+				InsecureSkipVerify: false},
+			errorMessage: "at most one of cert and cert_file must be configured",
+		},
+		{
+			configTLSConfig: TLSConfig{
+				CAFile:             "",
+				CertFile:           ClientCertificatePath,
+				Key:                Secret(readFile(t, ClientKeyNoPassPath)),
+				KeyFile:            ClientKeyNoPassPath,
+				ServerName:         "",
+				InsecureSkipVerify: false},
+			errorMessage: "at most one of key and key_file must be configured",
 		},
 	}
 
@@ -820,6 +900,32 @@ func TestBasicAuthPasswordFile(t *testing.T) {
 	}
 }
 
+func TestBasicUsernameFile(t *testing.T) {
+	cfg, _, err := LoadHTTPConfigFile("testdata/http.conf.basic-auth.username-file.good.yaml")
+	if err != nil {
+		t.Fatalf("Error loading HTTP client config: %v", err)
+	}
+	client, err := NewClientFromConfig(*cfg, "test")
+	if err != nil {
+		t.Fatalf("Error creating HTTP Client: %v", err)
+	}
+
+	rt, ok := client.Transport.(*basicAuthRoundTripper)
+	if !ok {
+		t.Fatalf("Error casting to basic auth transport, %v", client.Transport)
+	}
+
+	if rt.username != "" {
+		t.Errorf("Bad HTTP client username: %s", rt.username)
+	}
+	if string(rt.usernameFile) != "testdata/basic-auth-username" {
+		t.Errorf("Bad HTTP client usernameFile: %s", rt.usernameFile)
+	}
+	if string(rt.passwordFile) != "testdata/basic-auth-password" {
+		t.Errorf("Bad HTTP client passwordFile: %s", rt.passwordFile)
+	}
+}
+
 func getCertificateBlobs(t *testing.T) map[string][]byte {
 	files := []string{
 		TLSCAChainPath,
@@ -990,6 +1096,127 @@ func TestTLSRoundTripper(t *testing.T) {
 	}
 }
 
+func TestTLSRoundTripper_Inline(t *testing.T) {
+	handler := func(w http.ResponseWriter, r *http.Request) {
+		fmt.Fprint(w, ExpectedMessage)
+	}
+	testServer, err := newTestServer(handler)
+	if err != nil {
+		t.Fatal(err.Error())
+	}
+	defer testServer.Close()
+
+	testCases := []struct {
+		caText, caFile     string
+		certText, certFile string
+		keyText, keyFile   string
+
+		errMsg string
+	}{
+		{
+			// File-based everything.
+			caFile:   TLSCAChainPath,
+			certFile: ClientCertificatePath,
+			keyFile:  ClientKeyNoPassPath,
+		},
+		{
+			// Inline CA.
+			caText:   readFile(t, TLSCAChainPath),
+			certFile: ClientCertificatePath,
+			keyFile:  ClientKeyNoPassPath,
+		},
+		{
+			// Inline cert.
+			caFile:   TLSCAChainPath,
+			certText: readFile(t, ClientCertificatePath),
+			keyFile:  ClientKeyNoPassPath,
+		},
+		{
+			// Inline key.
+			caFile:   TLSCAChainPath,
+			certFile: ClientCertificatePath,
+			keyText:  readFile(t, ClientKeyNoPassPath),
+		},
+		{
+			// Inline everything.
+			caText:   readFile(t, TLSCAChainPath),
+			certText: readFile(t, ClientCertificatePath),
+			keyText:  readFile(t, ClientKeyNoPassPath),
+		},
+
+		{
+			// Invalid inline CA.
+			caText:   "badca",
+			certText: readFile(t, ClientCertificatePath),
+			keyText:  readFile(t, ClientKeyNoPassPath),
+
+			errMsg: "unable to use inline CA cert",
+		},
+		{
+			// Invalid cert.
+			caText:   readFile(t, TLSCAChainPath),
+			certText: "badcert",
+			keyText:  readFile(t, ClientKeyNoPassPath),
+
+			errMsg: "failed to find any PEM data in certificate input",
+		},
+		{
+			// Invalid key.
+			caText:   readFile(t, TLSCAChainPath),
+			certText: readFile(t, ClientCertificatePath),
+			keyText:  "badkey",
+
+			errMsg: "failed to find any PEM data in key input",
+		},
+	}
+
+	for i, tc := range testCases {
+		tc := tc
+		t.Run(strconv.Itoa(i), func(t *testing.T) {
+			cfg := HTTPClientConfig{
+				TLSConfig: TLSConfig{
+					CA:                 tc.caText,
+					CAFile:             tc.caFile,
+					Cert:               tc.certText,
+					CertFile:           tc.certFile,
+					Key:                Secret(tc.keyText),
+					KeyFile:            tc.keyFile,
+					InsecureSkipVerify: false},
+			}
+
+			c, err := NewClientFromConfig(cfg, "test")
+			if tc.errMsg != "" {
+				if !strings.Contains(err.Error(), tc.errMsg) {
+					t.Fatalf("Expected error message to contain %q, got %q", tc.errMsg, err)
+				}
+				return
+			} else if err != nil {
+				t.Fatalf("Error creating HTTP Client: %v", err)
+			}
+
+			req, err := http.NewRequest(http.MethodGet, testServer.URL, nil)
+			if err != nil {
+				t.Fatalf("Error creating HTTP request: %v", err)
+			}
+			r, err := c.Do(req)
+			if err != nil {
+				t.Fatalf("Can't connect to the test server")
+			}
+
+			b, err := io.ReadAll(r.Body)
+			r.Body.Close()
+			if err != nil {
+				t.Errorf("Can't read the server response body")
+			}
+
+			got := strings.TrimSpace(string(b))
+			if ExpectedMessage != got {
+				t.Errorf("The expected message %q differs from the obtained message %q", ExpectedMessage, got)
+			}
+		})
+	}
+}
+
 func TestTLSRoundTripperRaces(t *testing.T) {
 	bs := getCertificateBlobs(t)
 
@@ -1622,3 +1849,174 @@ func TestModifyTLSCertificates(t *testing.T) {
 		})
 	}
 }
+
+// loadHTTPConfigJSON parses the JSON input s into a HTTPClientConfig.
+func loadHTTPConfigJSON(buf []byte) (*HTTPClientConfig, error) {
+	cfg := &HTTPClientConfig{}
+	err := json.Unmarshal(buf, cfg)
+	if err != nil {
+		return nil, err
+	}
+	return cfg, nil
+}
+
+// loadHTTPConfigJSONFile parses the given JSON file into a HTTPClientConfig.
+func loadHTTPConfigJSONFile(filename string) (*HTTPClientConfig, []byte, error) {
+	content, err := os.ReadFile(filename)
+	if err != nil {
+		return nil, nil, err
+	}
+	cfg, err := loadHTTPConfigJSON(content)
+	if err != nil {
+		return nil, nil, err
+	}
+	return cfg, content, nil
+}
+
+func TestProxyConfig_Proxy(t *testing.T) {
+	var proxyServer *httptest.Server
+
+	defer func() {
+		if proxyServer != nil {
+			proxyServer.Close()
+		}
+	}()
+
+	proxyServerHandler := http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+		fmt.Fprintf(w, "Hello, %s", r.URL.Path)
+	})
+
+	proxyServer = httptest.NewServer(proxyServerHandler)
+
+	testCases := []struct {
+		name             string
+		proxyConfig      string
+		expectedProxyURL string
+		targetURL        string
+		proxyEnv         string
+		noProxyEnv       string
+	}{
+		{
+			name:             "proxy from environment",
+			proxyConfig:      `proxy_from_environment: true`,
+			expectedProxyURL: proxyServer.URL,
+			proxyEnv:         proxyServer.URL,
+			targetURL:        "http://prometheus.io/",
+		},
+		{
+			name:             "proxy_from_environment with no_proxy",
+			proxyConfig:      `proxy_from_environment: true`,
+			expectedProxyURL: "",
+			proxyEnv:         proxyServer.URL,
+			noProxyEnv:       "prometheus.io",
+			targetURL:        "http://prometheus.io/",
+		},
+		{
+			name:             "proxy_from_environment and localhost",
+			proxyConfig:      `proxy_from_environment: true`,
+			expectedProxyURL: "",
+			proxyEnv:         proxyServer.URL,
+			targetURL:        "http://localhost/",
+		},
+		{
+			name:             "valid proxy_url and localhost",
+			proxyConfig:      fmt.Sprintf(`proxy_url: %s`, proxyServer.URL),
+			expectedProxyURL: proxyServer.URL,
+			targetURL:        "http://localhost/",
+		},
+		{
+			name: "valid proxy_url and no_proxy and localhost",
+			proxyConfig: fmt.Sprintf(`proxy_url: %s
+no_proxy: prometheus.io`, proxyServer.URL),
+			expectedProxyURL: "",
+			targetURL:        "http://localhost/",
+		},
+		{
+			name:             "valid proxy_url",
+			proxyConfig:      fmt.Sprintf(`proxy_url: %s`, proxyServer.URL),
+			expectedProxyURL: proxyServer.URL,
+			targetURL:        "http://prometheus.io/",
+		},
+		{
+			name: "valid proxy url and no_proxy",
+			proxyConfig: fmt.Sprintf(`proxy_url: %s
+no_proxy: prometheus.io`, proxyServer.URL),
+			expectedProxyURL: "",
+			targetURL:        "http://prometheus.io/",
+		},
+		{
+			name: "valid proxy url and no_proxies",
+			proxyConfig: fmt.Sprintf(`proxy_url: %s
+no_proxy: promcon.io,prometheus.io,cncf.io`, proxyServer.URL),
+			expectedProxyURL: "",
+			targetURL:        "http://prometheus.io/",
+		},
+		{
+			name: "valid proxy url and no_proxies that do not include target",
+			proxyConfig: fmt.Sprintf(`proxy_url: %s
+no_proxy: promcon.io,cncf.io`, proxyServer.URL),
+			expectedProxyURL: proxyServer.URL,
+			targetURL:        "http://prometheus.io/",
+		},
+	}
+
+	for _, tc := range testCases {
+		t.Run(tc.name, func(t *testing.T) {
+			if proxyServer != nil {
+				defer proxyServer.Close()
+			}
+
+			var proxyConfig ProxyConfig
+
+			err := yaml.Unmarshal([]byte(tc.proxyConfig), &proxyConfig)
+			if err != nil {
+				t.Errorf("failed to unmarshal proxy config: %v", err)
+				return
+			}
+
+			if tc.proxyEnv != "" {
+				currentProxy := os.Getenv("HTTP_PROXY")
+				t.Cleanup(func() { os.Setenv("HTTP_PROXY", currentProxy) })
+				os.Setenv("HTTP_PROXY", tc.proxyEnv)
+			}
+
+			if tc.noProxyEnv != "" {
+				currentProxy := os.Getenv("NO_PROXY")
+				t.Cleanup(func() { os.Setenv("NO_PROXY", currentProxy) })
+				os.Setenv("NO_PROXY", tc.noProxyEnv)
+			}
+
+			req := httptest.NewRequest("GET", tc.targetURL, nil)
+
+			proxyFunc := proxyConfig.Proxy()
+			resultURL, err := proxyFunc(req)
+
+			if err != nil {
+				t.Fatalf("expected no error, but got: %v", err)
+				return
+			}
+			if tc.expectedProxyURL == "" && resultURL != nil {
+				t.Fatalf("expected no result URL, but got: %s", resultURL.String())
+				return
+			}
+			if tc.expectedProxyURL != "" && resultURL == nil {
+				t.Fatalf("expected result URL, but got nil")
+				return
+			}
+			if tc.expectedProxyURL != "" && resultURL.String() != tc.expectedProxyURL {
+				t.Fatalf("expected result URL: %s, but got: %s", tc.expectedProxyURL, resultURL.String())
+			}
+		})
+	}
+}
+
+func readFile(t *testing.T, filename string) string {
+	t.Helper()
+
+	content, err := os.ReadFile(filename)
+	if err != nil {
+		t.Fatalf("Failed to read file %q: %s", filename, err)
+	}
+
+	return string(content)
+}
